unit untSspiSrvClient;

(*******************************************************************************

Simle authentication client's class
For SSPIAuthServer.dpr
Author: Sergey N. Naberegnyh

Version 1.2.0.0
Created: November, 16, 2008
Updated: -  December, 03, 2008

*******************************************************************************)

interface

uses
  Windows, SysUtils, HPScktSrvr, untAuthSrvrMain,
  WinSock, JwaSspi, JwaWinError, untSharedTypes,
  HPSockApi;

type
  TSspiClientState = (SSPI_NOTCONNECTED, SSPI_DOAUTHENTICATE, SSPI_AUTHFAIL,
    SSPI_AUTHCOMPLETE, SSPI_WAITREQUEST, SSPI_SENDRESPONSE);

  TSspiSrvClient = class(THPServerClient)
  private
    FInitContext: boolean;
    FSecHandle: TSecHandle;
    FClientState: TSspiClientState;
    FDataWriteLen: LongInt;
    FBytesReaded: integer;
    FSecSizes: SecPkgContext_Sizes;
    FSignatureSize: DWORD;
    FUserName: AnsiString;
    FSignature: array[0..255] of Byte;
    procedure DoAuth();
    function SignBuffer(PData: PByte; DataLen: integer): boolean;
    function CheckSign(PData: PByte; DataLen: integer;
      out PHeader: PCmdHeader; out CmdLen: integer): boolean;
    procedure ProcessRequest(PData: PByte; DataLen: integer);
    procedure SendOK();
    procedure SendResponse(PBuf: Pointer; BufLen: integer);
    procedure ExecuteCmd_LIST(const Params: AnsiString);
    procedure ExecuteCmd_GET(const Params: AnsiString);
  protected
  public
    destructor Destroy; override;
    procedure Initialize;
    procedure Finalize;
    procedure DoReadComplete(BytesTransfered: Cardinal; CompletionKey: Integer);
    procedure DoWriteComplete(BytesTransfered: Cardinal; CompletionKey: Integer);
  end;

implementation

const
  sNotImpl_Template: AnsiString = 'User: %s'#13#10'Command: %s'#13#10 +
                                  'Status: OK - Not implemented'#13#10;
  sUnkCmd_Template:  AnsiString = 'User: %s'#13#10'Command: %s'#13#10 +
                                  'Status: ERROR - Unknown command'#13#10;
  sImpFail_Template: AnsiString = 'User: %s'#13#10'Command: %s'#13#10 +
                                  'Status: ERROR - Impersonation failed'#13#10;
  sOk_Template: AnsiString = 'User: %s; Connection OK'#13#10;                                  

{ TSspiSrvClient }

function TSspiSrvClient.CheckSign(PData: PByte; DataLen: integer;
  out PHeader: PCmdHeader; out CmdLen: integer): boolean;
var
  Desc: SecBufferDesc;
  SecBuf: array[0..1] of SecBuffer;
  QOP: Cardinal;
  SS: SECURITY_STATUS;
begin
  QOP:= 0;
  with Desc do
  begin
    ulVersion:= SECBUFFER_VERSION;
    cBuffers:= 2;
    pBuffers:= @SecBuf;
  end;

  with SecBuf[0] do
  begin
    cbBuffer:= PDWORD(PData)^;
    BufferType:= SECBUFFER_TOKEN;
    pvBuffer:= Pointer(Cardinal(PData) + SizeOf(DWORD));
  end;

  PHeader:= Pointer(Cardinal(PData) + SecBuf[0].cbBuffer + SizeOf(DWORD));
  CmdLen:= DataLen - integer(SecBuf[0].cbBuffer) - SizeOf(DWORD);
  with SecBuf[1] do
  begin
    cbBuffer:= CmdLen;
    BufferType:= SECBUFFER_DATA;
    pvBuffer:= PHeader;
  end;

  SS:= VerifySignature(@FSecHandle, @desc, 0, QOP);
  Result:= Succeeded(SS);
end;

destructor TSspiSrvClient.Destroy;
begin
  Finalize;
  inherited;
end;

procedure TSspiSrvClient.DoAuth;
var
  ss: SECURITY_STATUS;
  Expiry: TimeStamp;
  OutDesc, InDesc: SecBufferDesc;
  OutBuf, InBuf: SecBuffer;
  Attribs: ULONG;
  WsaBuf: array[0..1] of TWsaBuf;
  NameLen: Cardinal;
begin
  Attribs:= 0;

  with OutDesc do
  begin
    ulVersion:= SECBUFFER_VERSION;
    cBuffers:= 1;
    pBuffers:= @OutBuf;
  end;

  with OutBuf do
  begin
    cbBuffer:= BufferSize div 2;
    BufferType:= SECBUFFER_TOKEN;
    pvBuffer:= Buffer;
    Inc(PByte(pvBuffer), cbBuffer);
  end;

  with InDesc do
  begin
    ulVersion:= SECBUFFER_VERSION;
    cBuffers:= 1;
    pBuffers:= @InBuf;
  end;

  with InBuf do
  begin
    cbBuffer:= PDWORD(Buffer)^;
    BufferType:= SECBUFFER_TOKEN;
    pvBuffer:= Pointer(Cardinal(Buffer) + SizeOf(LongInt));
  end;

  if FInitContext then
    SS:= AcceptSecurityContext(@hCredHandle, @FSecHandle, @InDesc, Attribs,
         SECURITY_NATIVE_DREP, @FSecHandle, @OutDesc, Attribs,  @Expiry)
  else begin
    SS:= AcceptSecurityContext(@hCredHandle, nil, @InDesc, Attribs,
         SECURITY_NATIVE_DREP, @FSecHandle, @OutDesc, Attribs,  @Expiry);
    FInitContext:= Succeeded(SS);
  end;

  if Failed(SS) then
  begin
    FClientState:= SSPI_AUTHFAIL;
    Exit;
  end;

  if (SEC_I_COMPLETE_NEEDED = ss) or (SEC_I_COMPLETE_AND_CONTINUE = ss) then
  begin
    ss:= CompleteAuthToken(@FSecHandle, @OutDesc);
    if Failed(SS) then
    begin
      FClientState:= SSPI_AUTHFAIL;
      Exit;
    end;
  end;

  if SEC_E_OK = SS then
  begin
    ss:= QueryContextAttributes(@FSecHandle, SECPKG_ATTR_SIZES, @FSecSizes);
    if Failed(ss) then
    begin
      FClientState:= SSPI_AUTHFAIL;
      Exit;
    end;

    if Succeeded(ImpersonateSecurityContext(@FSecHandle)) then
    try
      NameLen:= SizeOf(FSignature);
      if GetUserName(@FSignature[0], NameLen) then
        SetString(FUserName, PAnsiChar(@FSignature[0]), NameLen - 1);
    finally
      if Failed(RevertSecurityContext(@FSecHandle)) then Disconnect;
    end;

    FClientState:= SSPI_AUTHCOMPLETE;
  end;

  if OutBuf.cbBuffer <> 0 then
  begin
    FDataWriteLen:= OutBuf.cbBuffer;
    WsaBuf[0].cLength:= sizeof(FDataWriteLen);
    WsaBuf[0].pBuffer:= @FDataWriteLen;
    WsaBuf[1].cLength:= OutBuf.cbBuffer;
    WsaBuf[1].pBuffer:= OutBuf.pvBuffer;
    if 0 <> Write(WsaBuf[0], 2, 0) then FClientState:= SSPI_AUTHFAIL;
  end;
end;

procedure TSspiSrvClient.DoReadComplete(BytesTransfered: Cardinal;
  CompletionKey: Integer);
var
  WsaBuf: TWsaBuf;
  DataLen: longint;
  MaxData: integer;
begin
  if FClientState = SSPI_DOAUTHENTICATE then
    MaxData:= BufferSize div 2 else
    MaxData:= BufferSize;

  Inc(FBytesReaded, BytesTransfered);
  if FBytesReaded >= SizeOf(LongInt) then
  begin
    DataLen:= PLongInt(Buffer)^;
    if DataLen > MaxData then
    begin
      Disconnect;
      Exit;
    end;

    if FBytesReaded < DataLen + SizeOf(LongInt) then
    begin
      WsaBuf.cLength:= DataLen - (FBytesReaded - SizeOf(LongInt));
      WsaBuf.pBuffer:= Buffer;
      Inc(WsaBuf.pBuffer, FBytesReaded);
      if 0 <> Read(WsaBuf, 1, CompletionKey + 1) then Disconnect;
      Exit;
    end else
    begin
      case FClientState of

        SSPI_DOAUTHENTICATE:
        begin
          FBytesReaded:= 0;
          DoAuth;
          case FClientState of
            SSPI_AUTHFAIL:
            begin
              FDataWriteLen:= 0;
              WsaBuf.cLength:= sizeof(FDataWriteLen);
              WsaBuf.pBuffer:= @FDataWriteLen;
              Write(WsaBuf, 1, 0);
              Disconnect;
            end;

            SSPI_AUTHCOMPLETE: SendOK;
          end;
        end;

        SSPI_WAITREQUEST:
          ProcessRequest(PByte(Cardinal(Buffer) + SizeOf(LongInt)), PLongInt(Buffer)^);

        else Disconnect;
      end;
    end;
  end else
  begin
    WsaBuf.cLength:= MaxData - FBytesReaded;
    WsaBuf.pBuffer:= Buffer;
    Inc(WsaBuf.pBuffer, FBytesReaded);
    if 0 <> Read(WsaBuf, 1, 0) then Disconnect;
  end;
end;

procedure TSspiSrvClient.DoWriteComplete(BytesTransfered: Cardinal;
  CompletionKey: Integer);
var
  WsaBuf: TWsaBuf;
begin
  if not Connected then Exit;
  if FClientState = SSPI_DOAUTHENTICATE then
    WsaBuf.cLength:= BufferSize div 2
  else
    WsaBuf.cLength:= BufferSize;
  WsaBuf.pBuffer:= Buffer;
  if FClientState in [SSPI_AUTHCOMPLETE, SSPI_SENDRESPONSE] then
    FClientState:= SSPI_WAITREQUEST;
  if 0 <> Read(WsaBuf, 1, 0) then Disconnect;
end;

procedure TSspiSrvClient.ExecuteCmd_GET(const Params: AnsiString);
var
  s: AnsiString;
begin
  if Succeeded(ImpersonateSecurityContext(@FSecHandle)) then
  try
  // replace this code to your implementation

    s:= Format(sNotImpl_Template, [FUserName, 'GET', 0]);
    StrPCopy(PAnsiChar(Buffer), s);
  finally
   if Failed(RevertSecurityContext(@FSecHandle)) then Disconnect;
  end else
  begin
    s:= Format(sImpFail_Template, [FUserName, 'GET']);
    StrPCopy(PAnsiChar(Buffer), s);
  end;
  SendResponse(Buffer, Length(s));
end;

procedure TSspiSrvClient.ExecuteCmd_LIST(const Params: AnsiString);
var
  s: AnsiString;
begin
  if Succeeded(ImpersonateSecurityContext(@FSecHandle)) then
  try
  // replace this code to your implementation

    s:= Format(sNotImpl_Template, [FUserName, 'LIST', 0]);
    StrPCopy(PAnsiChar(Buffer), s);
  finally
   if Failed(RevertSecurityContext(@FSecHandle)) then Disconnect;
  end else
  begin
    s:= Format(sImpFail_Template, [FUserName, 'LIST']);
    StrPCopy(PAnsiChar(Buffer), s);
  end;
  SendResponse(Buffer, Length(s));
end;

procedure TSspiSrvClient.Finalize;
begin
  if FInitContext then
  begin
    FInitContext:= false;
    DeleteSecurityContext(@FSecHandle);
//    FillChar(FSecHandle, SizeOf(FSecHandle), 0);
  end;
  FClientState:= SSPI_NOTCONNECTED;
end;

procedure TSspiSrvClient.Initialize;
begin
  FClientState:= SSPI_DOAUTHENTICATE;
end;

procedure TSspiSrvClient.ProcessRequest(PData: PByte; DataLen: integer);
var
  PHead: PCmdHeader;
  PCmd: PAnsiChar;
  PParam: PAnsiChar;
  sCmd, sParam: AnsiString;
  CmdLen: integer;
  s: string;
begin
  if not CheckSign(PData, DataLen, PHead, CmdLen) then
  begin
    Disconnect;
    Exit;
  end;

  PCmd:= Pointer(PHead);
  Inc(PCmd, PHead.CmdOffset);
  SetString(sCmd, PCmd, PHead.CmdLen);
  PParam:= Pointer(PHead);
  Inc(PParam, PHead.ParamOffset);
  SetString(sParam, PParam, PHead.ParamLen);

  if sCmd = 'LIST' then ExecuteCmd_LIST(sParam) else
  if sCmd = 'GET'  then ExecuteCmd_GET(sParam)  else
  begin
    s:= Format(sUnkCmd_Template, [FUserName, sCmd]);
    StrPCopy(PAnsiChar(Buffer), s);
    SendResponse(Buffer, Length(s));
  end;
end;

procedure TSspiSrvClient.SendOK;
var
  S: string;
begin
  s:= Format(sOk_Template, [FUserName]);
  StrPCopy(PAnsiChar(Buffer), s);
  SendResponse(Buffer, Length(s));
end;

procedure TSspiSrvClient.SendResponse(PBuf: Pointer; BufLen: integer);
var
  WsaBuf: array[0..3] of TWsaBuf;
begin
  if SignBuffer(PBuf, BufLen) then
  begin
{
    if BadFlag then
    begin
      BadFlag:= false;
      PAnsiChar(PBuf)^:= 'A';
    end;
}
    WsaBuf[1].cLength:= SizeOf(FSignatureSize);
    WsaBuf[1].pBuffer:= @FSignatureSize;
    WsaBuf[2].cLength:= FSignatureSize;
    WsaBuf[2].pBuffer:= @FSignature[0];
    WsaBuf[3].cLength:= BufLen;
    WsaBuf[3].pBuffer:= PBuf;
    FDataWriteLen:= wsabuf[1].cLength + WsaBuf[2].cLength + WsaBuf[3].cLength;
    WsaBuf[0].cLength:= SizeOf(FDataWriteLen);
    WsaBuf[0].pBuffer:= @FDataWriteLen;
    FClientState:= SSPI_SENDRESPONSE;
    if 0 <> Write(WsaBuf[0], 4, 0) then Disconnect;
  end else
    Disconnect;
end;

function TSspiSrvClient.SignBuffer(PData: PByte;
  DataLen: integer): boolean;
var
  Desc: SecBufferDesc;
  SecBuf: array[0..2] of SecBuffer;
  SS: SECURITY_STATUS;
begin
  with Desc do
  begin
    ulVersion:= SECBUFFER_VERSION;
    cBuffers:= 3;
    pBuffers:= @SecBuf;
  end;

  with SecBuf[0] do
  begin
    cbBuffer:= FSecSizes.cbMaxSignature;
    BufferType:= SECBUFFER_TOKEN;
    pvBuffer:= @FSignature;
  end;

  with SecBuf[1] do
  begin
    cbBuffer:= DataLen;
    BufferType:= SECBUFFER_DATA;
    pvBuffer:= PData;
  end;

  with SecBuf[2] do
  begin
    cbBuffer:= 0;
    BufferType:= SECBUFFER_PADDING;
    pvBuffer:= nil;
  end;

  SS:= MakeSignature(@FSecHandle, 0, @Desc, 0);
  Result:= Succeeded(SS);
  if Result then FSignatureSize:= SecBuf[0].cbBuffer;
end;

end.
