unit untSslClient;

(*******************************************************************************

Author: Sergey N. Naberegnyh

Created: November, 11, 2008
Updated: December, 03, 2008

*******************************************************************************)

interface

uses
  Windows, Messages, SysUtils, Classes,
  SockSSPIMics, HPSockApi, HPScktSrvr,
  WinSock, untClient, JwaWinError, JwaWinCrypt,
  JwaSspi;

type
  TSSLClientState = (SSL_NOTCONNECTED, SSL_DONEGOTIATE, SSL_NEGOTIATEFAIL,
    SSL_NEGOTIATECOMPLETE, SSL_WAITREQUEST, SSL_SENDHEAD, SSL_SENDBODY);

  ESslException = class(Exception);
  ESslCertException = class(ESslException);
  ESslCertStoreException = class(ESslException);

  THttpSrvClientSSL = class(THttpSrvClient)
  private
    FClientState: TSSLClientState;
    FInitContext: boolean;
    FContextAccepted: integer;
    FContext: CtxtHandle;
    FBytesReaded: integer;
    FOutBuffer: SecBuffer;
    FSizes: SecPkgContext_StreamSizes;
    FSslHeader, FSslTrailer: array[0..1, 0..31] of Byte;
    FBytesSent: integer;
    FSSLConnected: integer;
    FExtraDataSize: integer;

    procedure DoNegotiate();
    procedure DecryptBuffer();
    procedure FreeSecurityContext();
  protected
    procedure SendResponse(PData: Pointer; DataLen: integer); override;
  public
    procedure Initialize; override;
    procedure Finalize; override;
    procedure DoWriteComplete(BytesTransfered: Cardinal;
      CompletionKey, Error: Integer); override;
    procedure ProcessBuffer(PBuf: PAnsiChar; BufLen: integer); override;
    procedure SSLDisconnect();

    property ClientState: TSSLClientState read FClientState;
  end;

  TSSL_Security = class
  private
    FSecurityLib: HMODULE;
    FpSSPI: PSecurityFunctionTable;
    FCredHandle: CredHandle;
    FMyCertStore: HCERTSTORE;
    FMutualAuth: boolean;
    procedure LoadSecurityLibrary();
    procedure GetCredentials(const CertName: AnsiString;
      MachineStore: boolean; Protocols: Cardinal);
  public
    constructor Create(const UserNmae: AnsiString;
      MachineStore, MutualAuth: boolean; Protocols: Cardinal);
    destructor Destroy; override;

    property SSPI: PSecurityFunctionTable read FpSSPI;
    property MutualAuthenticate: boolean read FMutualAuth;
  end;

  TByteArray = array[0..MAXINT - 1] of byte;
  PByteArray = ^TByteArray;

const
  IO_BUFFER_SIZE = $10000;

var
  SslSecurity: TSSL_Security = nil;
    
implementation

{ TSSL_Security }

constructor TSSL_Security.Create(const UserNmae: AnsiString;
  MachineStore, MutualAuth: boolean; Protocols: Cardinal);
begin
  if (Win32Platform <> VER_PLATFORM_WIN32_NT) or not CheckWin32Version(5, 1)
  then raise ESslException.Create(SysErrorMessage(ERROR_OLD_WIN_VERSION));
  FMutualAuth:= MutualAuth;
  LoadSecurityLibrary;
  GetCredentials(UserNmae, MachineStore, Protocols);
end;

destructor TSSL_Security.Destroy;
begin
  if 0 <> FSecurityLib then FreeLibrary(FSecurityLib);
  if FMyCertStore <> nil then CertCloseStore(FMyCertStore, 0);
  if (FCredHandle.dwLower or FCredHandle.dwUpper <> 0) then
    if nil <> SSPI then SSPI.FreeCredentialHandle(@FCredHandle);
  inherited;
end;

procedure TSSL_Security.GetCredentials(const CertName: AnsiString;
  MachineStore: boolean; Protocols: Cardinal);
var
  NameLen: Cardinal;
  sName: AnsiString;
var
  SChannelCred: SCHANNEL_CRED;
  Expiry: TimeStamp;
  ss: SECURITY_STATUS;
  pCertContext: PCCERT_CONTEXT;
begin
  sName:= CertName;
  UniqueString(sName);
  if sName = '' then
  begin
    NameLen:= 256;
    SetLength(sName, NameLen);
    if GetUserName(@sName[1], NameLen) then
      SetLength(sName, NameLen) else sName:= '';
  end;
  if sName = '' then
    raise ESslCertException.Create('Error getting user nsme');

  if MachineStore then
    FMyCertStore:= CertOpenStore(CERT_STORE_PROV_SYSTEM,
            X509_ASN_ENCODING, 0, CERT_SYSTEM_STORE_LOCAL_MACHINE, PAnsiChar('MY'))
  else
    FMyCertStore:= CertOpenSystemStore(0, 'MY');
  if nil = FMyCertStore then raise ESslCertStoreException.CreateFmt(
    'Error opening storage of certificates: %s', [SysErrorMessage(GetLastError)]);


  pCertContext:= CertFindCertificateInStore(FMyCertStore, X509_ASN_ENCODING,
                              0, CERT_FIND_SUBJECT_STR_A, PAnsiChar(sName), nil);
  if pCertContext = nil then raise ESslCertException.CreateFmt(
    'Certificate not found: %s', [SysErrorMessage(GetLastError)]);

  try
    FillChar(SChannelCred, sizeof(SChannelCred), 0);
    with SChannelCred do
    begin
      dwVersion:= SCHANNEL_CRED_VERSION;
      cCreds:= 1;
      paCred:= @pCertContext;
      grbitEnabledProtocols:= Protocols;;
    end;
    ss:= SSPI.AcquireCredentialsHandleA(nil, @UNISP_NAME_A[1],
      SECPKG_CRED_INBOUND, nil, @SchannelCred, nil, nil, @FCredHandle, @Expiry);
    if ss <> SEC_E_OK then  raise ESslCertException.CreateFmt(
      'Error $%x returned by AcquireCredentialsHandle', [ss]);
      
  finally
    CertFreeCertificateContext(pCertContext);
  end;
end;

procedure TSSL_Security.LoadSecurityLibrary;
var
  pInitIntf: INIT_SECURITY_INTERFACE;
begin
  FSecurityLib:= LoadLibrary(DLL_NAME);
  if FSecurityLib = 0 then
    raise ESslException.CreateFmt('Error $%x loading security library',
      [GetLastError]);

  @pInitIntf:= GetProcAddress(FSecurityLib, 'InitSecurityInterfaceA');
  if @pInitIntf = nil then
    raise ESslException.CreateFmt(
      'Error $%x getting InitSecurityInterface entry point.', [GetLastError()]);

  FpSSPI:= pInitIntf();
  if FpSSPI = nil then raise ESslException.CreateFmt(
      'Error $%x reading security interface.', [GetLastError()]);
end;

{ THttpSrvClientSSL }

procedure THttpSrvClientSSL.DecryptBuffer();
var
  ClientMsg: SecBufferDesc;
  Buffers: array[0..3] of SecBuffer;
  SS: SECURITY_STATUS;
  n: integer;
  WsaBuf: TWsaBuf;
begin
  with ClientMsg do
  begin
    ulVersion:= SECBUFFER_VERSION;
    cBuffers:= 4;
    pBuffers:= @Buffers;
  end;  

  if FExtraDataSize > 0 then
  begin
    Move(PByteArray(Buffer)[FSizes.cbMaximumMessage], Buffer^, FExtraDataSize);
    FBytesReaded:= FExtraDataSize;
    FExtraDataSize:= 0;
  end;

  with Buffers[0] do
  begin
    pvBuffer:= Buffer;
    cbBuffer:= FBytesReaded;
    BufferType:= SECBUFFER_DATA;
  end;

  Buffers[1].BufferType:= SECBUFFER_EMPTY;
  Buffers[2].BufferType:= SECBUFFER_EMPTY;
  Buffers[3].BufferType:= SECBUFFER_EMPTY;

  SS:= SslSecurity.SSPI.DecryptMessage(@FContext, @ClientMsg, 0, Cardinal(nil^));
  case SS of
    SEC_E_OK:
    begin
{$IFDEF LOGGING_ON}
      Server.LogMessage(['DecryptBuffer - OK'], 0, 0, Cardinal(Self));
{$ENDIF}

      FBytesReaded:= 0;
      FExtraDataSize:= 0;
      
      n:= High(Buffers);
      repeat
        if Buffers[n].BufferType = SECBUFFER_EXTRA then
        begin
          Move(Buffers[n].pvBuffer^,
            PByteArray(Buffer)[FSizes.cbMaximumMessage], Buffers[n].cbBuffer);
          FExtraDataSize:= Buffers[n].cbBuffer;
        end else
          if Buffers[n].BufferType = SECBUFFER_DATA then
          begin
            FBytesSent:= 0;
            inherited ProcessBuffer(PAnsiChar(Buffers[n].pvBuffer), Buffers[n].cbBuffer);
            Break;
          end;
        Dec(n);
      until n < 1;

    end;

    SEC_E_INCOMPLETE_MESSAGE:
    begin
{$IFDEF LOGGING_ON}
      Server.LogMessage(['DecryptBuffer - Incomplete'], 0, 0, Cardinal(Self));
{$ENDIF}
      WsaBuf.cLength:= BufferSize - FBytesReaded;
      WsaBuf.pBuffer:= Buffer;
      Inc(WsaBuf.pBuffer, FBytesReaded);
      if 0 <> Read(WsaBuf, 1, 0) then SSLDisconnect;
    end;

    else
    begin
      FBytesReaded:= 0;
      {$IFDEF LOGGING_ON}Server.LogMessage(['DecryptBuffer - failed !!!!!!', PAnsiChar(IntToHex(ss, 0))], 0, 0, Cardinal(Self));{$ENDIF}
      SSLDisconnect;
    end;
  end;
end;

procedure THttpSrvClientSSL.DoNegotiate();
var
  SS: SECURITY_STATUS;
  Expiry:     TimeStamp;
  InDesc:   SecBufferDesc;
  OutDesc:  SecBufferDesc;
  InBuffers:  array[0..1] of SecBuffer;
  SSPIFlags, SSPIOutFlags: DWORD;
  WsaBuf: TWsaBuf;
  Writting: boolean;
  pRemoteCertContext: PCCERT_CONTEXT;
begin
  SSPIFlags:= ASC_REQ_SEQUENCE_DETECT or ASC_REQ_REPLAY_DETECT  or
              ASC_REQ_CONFIDENTIALITY or ASC_REQ_EXTENDED_ERROR or
              ASC_REQ_ALLOCATE_MEMORY or ASC_REQ_STREAM;

  if SslSecurity.MutualAuthenticate then
    SSPIFlags:= SSPIFlags or ASC_REQ_MUTUAL_AUTH;

  Writting:= false;
  with OutDesc do
  begin
    cBuffers:= 1;
    pBuffers:= @FOutBuffer;
    ulVersion:= SECBUFFER_VERSION;
  end;

  repeat

    with InBuffers[0] do
    begin
      pvBuffer:= Buffer;
      InBuffers[0].cbBuffer:= FBytesReaded;
      InBuffers[0].BufferType:= SECBUFFER_TOKEN;
    end;

    with InBuffers[1] do
    begin
      pvBuffer  := nil;
      cbBuffer  := 0;
      BufferType:= SECBUFFER_EMPTY;
    end;

    with InDesc do
    begin
      cBuffers       := 2;
      pBuffers       := @InBuffers;
      ulVersion      := SECBUFFER_VERSION;
    end;

    with FOutBuffer do
    begin
      pvBuffer  := nil;
      BufferType:= SECBUFFER_TOKEN;
      cbBuffer  := 0;
    end;

    with SslSecurity do
      if FInitContext then
      begin
        SS:= SSPI.AcceptSecurityContext(
            @FCredHandle, nil, @InDesc, SSPIFlags, SECURITY_NATIVE_DREP,
            @FContext, @OutDesc, SSPIOutFlags, @Expiry);
        FInitContext:= false;
        if (FContext.dwLower or FContext.dwUpper) <> 0 then
          InterlockedExchange(FContextAccepted, 1);
      end else
        SS:= SSPI.AcceptSecurityContext(
            @FCredHandle, @FContext, @InDesc, SSPIFlags,
            SECURITY_NATIVE_DREP, nil, @OutDesc, SSPIOutFlags, @Expiry);

    case SS of
      SEC_E_OK, SEC_I_CONTINUE_NEEDED:
      begin
        if SS = SEC_E_OK then
        begin
          SslSecurity.SSPI.QueryContextAttributesA(@FContext,
                                     SECPKG_ATTR_STREAM_SIZES, @FSizes);

          if SslSecurity.MutualAuthenticate then
          begin
            ss:= SslSecurity.SSPI.QueryContextAttributesA(@FContext,
                  SECPKG_ATTR_REMOTE_CERT_CONTEXT, @pRemoteCertContext);
            if SEC_E_OK = ss then
            begin
              // Verify client certificate etc.
              //...
              CertFreeCertificateContext(pRemoteCertContext);
            end else
            begin
              FreeSecurityContext;
              FClientState:= SSL_NEGOTIATEFAIL;
              Exit;
            end;
          end;

          FClientState:= SSL_NEGOTIATECOMPLETE;
          InterlockedExchange(FSSLConnected, 1);
        end;

        if InBuffers[1].BufferType = SECBUFFER_EXTRA then
        begin
          Move(PByteArray(Buffer)[FBytesReaded - integer(InBuffers[1].cbBuffer)],
                                     Buffer^, InBuffers[1].cbBuffer);
          FBytesReaded:= InBuffers[1].cbBuffer;
        end else
          FBytesReaded:= 0;

        if FOutBuffer.pvBuffer <> nil then
        begin
          WsaBuf.cLength:= FOutBuffer.cbBuffer;
          WsaBuf.pBuffer:= FOutBuffer.pvBuffer;
          if 0 <> Write(WsaBuf, 1, 0) then
          begin
            SslSecurity.SSPI.FreeContextBuffer(FOutBuffer.pvBuffer);
            FOutBuffer.pvBuffer:= nil;
            SSLDisconnect;
            FClientState:= SSL_NEGOTIATEFAIL;
            Exit;
          end;
          Writting:= true;
        end else
          if SS = SEC_E_OK then
          begin
            FClientState:= SSL_WAITREQUEST;
            if FBytesReaded = 0 then
            begin
              WsaBuf.cLength:= BufferSize;
              WsaBuf.pBuffer:= Buffer;
              ss:= Read(WsaBuf, 1, 0);
              if ss <> 0 then
              begin
{$IFDEF LOGGING_ON}
                Server.LogMessage(['DoNegotiate - Negotiate failed on Read !!!!! - ',
                  PAnsiChar('$' + IntToHex(SS, 8))], 0, 0, Cardinal(Self));
{$ENDIF}
                SSLDisconnect;
              end;
            end;
          end;

{$IFDEF LOGGING_ON}
          if SS = SEC_I_CONTINUE_NEEDED then
            Server.LogMessage(['DoNegotiate - Continue Needed'], 0, 0, Cardinal(Self))
          else
            Server.LogMessage(['DoNegotiate - Negotiate complete'], 0, 0, Cardinal(Self));
{$ENDIF}

        if (SS = SEC_E_OK) or (FBytesReaded = 0) then Exit;
      end; // SEC_E_OK, SEC_I_CONTINUE_NEEDED

      SEC_E_INCOMPLETE_MESSAGE:
      begin
{$IFDEF LOGGING_ON}
        Server.LogMessage(['DoNegotiate - Incomplete message'], 0, 0, Cardinal(Self));
{$ENDIF}
        if Writting then Exit;
        WsaBuf.cLength:= BufferSize - FBytesReaded;
        WsaBuf.pBuffer:= Buffer;
        Inc(WsaBuf.pBuffer, FBytesReaded);
        if 0 <> Read(WsaBuf, 1, 0) then FClientState:= SSL_NEGOTIATEFAIL;
{$IFDEF LOGGING_ON}
        if ClientState = SSL_NEGOTIATEFAIL then
          Server.LogMessage(['DoNegotiate - Negotiate failed on Read !!!!!'], 0, 0, Cardinal(Self));
{$ENDIF}
        Exit;
      end; // SEC_E_INCOMPLETE_MESSAGE

      else begin
        FClientState:= SSL_NEGOTIATEFAIL;
{$IFDEF LOGGING_ON}
        Server.LogMessage(['DoNegotiate - Negotiate failed !!!!!'], 0, 0, Cardinal(Self));
{$ENDIF}
        if FAILED(SS) and (0 <> (SSPIOutFlags and ASC_RET_EXTENDED_ERROR)) then
        begin
          WsaBuf.cLength:= FOutBuffer.cbBuffer;
          WsaBuf.pBuffer:= FOutBuffer.pvBuffer;
          if 0 <> Write(WsaBuf, 1, 0) then
          begin
            SslSecurity.SSPI.FreeContextBuffer(FOutBuffer.pvBuffer);
            FOutBuffer.pvBuffer:= nil;
          end;
        end;
        Exit;
      end;

    end;  // case SS of
  until false;
end;

procedure THttpSrvClientSSL.DoWriteComplete(BytesTransfered: Cardinal;
  CompletionKey, Error: Integer);
var
  Wsabuf: TWsaBuf;
begin
  if FOutBuffer.pvBuffer <> nil then
  begin
    SslSecurity.SSPI.FreeContextBuffer(FOutBuffer.pvBuffer);
    FOutBuffer.pvBuffer:= nil;
  end;

  case ClientState of
    SSL_DONEGOTIATE:
    begin
      Wsabuf.cLength:= BufferSize - FBytesReaded;
      Wsabuf.pBuffer:= Buffer;
      Inc(Wsabuf.pBuffer, FBytesReaded);
      if 0 <> Read(Wsabuf, 1, 0) then
      begin
        KeepAlive:= false;
        inherited;
        Disconnect;
      end;
    end;

    SSL_NEGOTIATECOMPLETE:
    begin
      Wsabuf.cLength:= BufferSize - FBytesReaded;
      Wsabuf.pBuffer:= Buffer;
      Inc(Wsabuf.pBuffer, FBytesReaded);
      FClientState:= SSL_WAITREQUEST;
      if 0 <> Read(Wsabuf, 1, 0) then
      begin
        KeepAlive:= false;
        inherited;
        SSLDisconnect;
      end;
    end;

    SSL_NEGOTIATEFAIL:
    begin
      inherited;
      Disconnect;
    end;

    SSL_SENDHEAD,
    SSL_SENDBODY:  SendResponse(nil, 0);
  end;
end;

procedure THttpSrvClientSSL.Finalize;
begin
  FClientState:= SSL_NOTCONNECTED;
  SSLDisconnect;
  FreeSecurityContext;
  inherited;
end;

procedure THttpSrvClientSSL.FreeSecurityContext;
begin
  if 0 <> InterlockedExchange(FContextAccepted, 0) then
  begin
    SslSecurity.SSPI.DeleteSecurityContext(@Fcontext);
    FillChar(FContext, SizeOf(FContext), 0);
  end;
end;

procedure THttpSrvClientSSL.Initialize;
begin
  inherited;
  FClientState:= SSL_DONEGOTIATE;
  FExtraDataSize:= 0;
  FInitContext:= true;
  FBytesReaded:= 0;
  FAuthorized:= true;
end;

procedure THttpSrvClientSSL.ProcessBuffer(PBuf: PAnsiChar;
  BufLen: integer);
begin
{$IFDEF LOGGING_ON}
  Server.LogMessage(['ProcessBuffer_Ex - IN'], 0, 0, Cardinal(Self));
{$ENDIF}
  if BufLen = 0 then
  begin
    if ClientState > SSL_NEGOTIATEFAIL then SSLDisconnect;
    Exit;
  end;

  Inc(FBytesReaded, BufLen);
  if FClientState = SSL_DONEGOTIATE then
  begin
    DoNegotiate();
    if ClientState = SSL_NEGOTIATEFAIL then
    begin
      Disconnect;
      Exit;
    end else
      if ClientState = SSL_NEGOTIATECOMPLETE then Exit;
  end;

  case ClientState of
    SSL_NOTCONNECTED:      Disconnect;
    SSL_DONEGOTIATE:       Exit;
    SSL_NEGOTIATEFAIL:     Disconnect;
    SSL_WAITREQUEST:       DecryptBuffer();
    else                   SSLDisconnect;
  end;
end;

procedure THttpSrvClientSSL.SendResponse(PData: Pointer; DataLen: integer);
var
  SrvrMessage: SecBufferDesc; 
  Buffers: array[0..7] of SecBuffer;
  SS: SECURITY_STATUS;
  BufCount: integer;
  WsaBuf: array[0..5] of TWsaBuf;
  BytesRead: Cardinal;
begin
{$IFDEF LOGGING_ON}
  if PData <> nil then
    Server.LogMessage(['SendResponse - IN'#13#10, PAnsiChar(PData)], 0, 0, Cardinal(Self))
  else
    Server.LogMessage(['SendResponse - IN - NIL'], 0, 0, Cardinal(Self));
{$ENDIF}

  BufCount:= 0;
  if PData <> nil then
  begin
    FClientState:= SSL_SENDHEAD;
    with SrvrMessage do
    begin
      ulVersion:= SECBUFFER_VERSION;
      cBuffers:= 4;
      pBuffers:= @Buffers[0];
    end;

//    FillChar(FSslHeader[0], SizeOf(FSslHeader[0]), 0);
    with Buffers[0] do
    begin
      pvBuffer:= @FSslHeader[0];
      cbBuffer:= FSizes.cbHeader;
      BufferType:= SECBUFFER_STREAM_HEADER;
    end;

    with Buffers[1] do
    begin
      pvBuffer:= PData;
      cbBuffer:= DataLen;
      BufferType:= SECBUFFER_DATA;
    end;

    with Buffers[2] do
    begin
      pvBuffer:= @FSslTrailer[0];
      cbBuffer:= FSizes.cbTrailer;
      BufferType:= SECBUFFER_STREAM_TRAILER;
    end;

    Buffers[3].BufferType:= SECBUFFER_EMPTY;

    SS:= SslSecurity.SSPI.EncryptMessage(@FContext, 0, @SrvrMessage, 0);
    if FAILED(SS) then
    begin
{$IFDEF LOGGING_ON}
      Server.LogMessage(['SendResponse - buf encrypt failed', PAnsiChar(IntToHex(ss, 0))], 0, 0, Cardinal(Self));
{$ENDIF}
      SSLDisconnect;
      Exit;
    end;
{$IFDEF LOGGING_ON}
    Server.LogMessage(['SendResponse - buf encrypt success'], 0, 0, Cardinal(Self));
{$ENDIF}

    WsaBuf[0].cLength:= Buffers[0].cbBuffer;
    WsaBuf[0].pBuffer:= Buffers[0].pvBuffer;
    WsaBuf[1].cLength:= Buffers[1].cbBuffer;
    WsaBuf[1].pBuffer:= Buffers[1].pvBuffer;
    WsaBuf[2].cLength:= Buffers[2].cbBuffer;
    WsaBuf[2].pBuffer:= Buffers[2].pvBuffer;

    Inc(BufCount);

  end;

  FClientState:= SSL_SENDBODY;
  BytesRead:= 0;
  if FileHandle <> 0 then
  begin
    if ReadFile(FileHandle, Buffer^, FSizes.cbMaximumMessage, BytesRead, nil) then
    begin
      if BytesRead > 0 then
      begin
        with SrvrMessage do
        begin
          ulVersion:= SECBUFFER_VERSION;
          cBuffers:= 4;
          pBuffers:= @Buffers[4];
        end;

//        FillChar(FSslHeader[1], SizeOf(FSslHeader[1]), 0);
        with Buffers[4] do
        begin
          pvBuffer:= @FSslHeader[1];
          cbBuffer:= FSizes.cbHeader;
          BufferType:= SECBUFFER_STREAM_HEADER;
        end;

        with Buffers[5] do
        begin
          pvBuffer:= Buffer;
          cbBuffer:= BytesRead;
          BufferType:= SECBUFFER_DATA;
        end;

        with Buffers[6] do
        begin
          pvBuffer:= @FSslTrailer[1];
          cbBuffer:= FSizes.cbTrailer;
          BufferType:= SECBUFFER_STREAM_TRAILER;
        end;

        Buffers[7].BufferType:= SECBUFFER_EMPTY;

        SS:= SslSecurity.SSPI.EncryptMessage(@FContext, 0, @SrvrMessage, 0);
        if FAILED(SS) then
        begin
          SSLDisconnect;
          Exit;
        end;
{$IFDEF LOGGING_ON}
        Server.LogMessage(['SendResponse - File encrypt success'], 0, 0, Cardinal(Self));
{$ENDIF}

        WsaBuf[BufCount * 3].cLength:=     Buffers[4].cbBuffer;
        WsaBuf[BufCount * 3].pBuffer:=     Buffers[4].pvBuffer;
        WsaBuf[BufCount * 3 + 1].cLength:= Buffers[5].cbBuffer;
        WsaBuf[BufCount * 3 + 1].pBuffer:= Buffers[5].pvBuffer;
        WsaBuf[BufCount * 3 + 2].cLength:= Buffers[6].cbBuffer;
        WsaBuf[BufCount * 3 + 2].pBuffer:= Buffers[6].pvBuffer;

        Inc(BufCount);
      end;
    end else
      KeepAlive:= false;
    if BytesRead < FSizes.cbMaximumMessage then CloseFile;
  end;

  if BufCount > 0 then
  begin
{$IFDEF LOGGING_ON}
    Server.LogMessage(['SendResponse - Send OK', PAnsiChar(IntToStr(BufCount))], 0, 0, Cardinal(Self));
{$ENDIF}
    ss:= Write(WsaBuf[0], BufCount * 3, 0);
    if 0 <> ss then
    begin
{$IFDEF LOGGING_ON}
      Server.LogMessage(['SendResponse - Send failed !!!!!', PAnsiChar(IntToStr(ss))], 0, 0, Cardinal(Self));
{$ENDIF}
      SSLDisconnect;
    end;
  end else
  begin
    if not KeepAlive then SSLDisconnect else
    begin
      FClientState:= SSL_WAITREQUEST;
      if FExtraDataSize > 0 then
      begin
{$IFDEF LOGGING_ON}
        Server.LogMessage(['SendResponse - Extra data', PAnsiChar(IntToStr(FExtraDataSize))], 0, 0, Cardinal(Self));
{$ENDIF}
        DecryptBuffer;
        Exit;
      end;
      WsaBuf[0].cLength:= BufferSize;
      WsaBuf[0].pBuffer:= Buffer;
      ss:= Read(WsaBuf[0], 1, 0);
      if 0 <> ss then
      begin
{$IFDEF LOGGING_ON}
        Server.LogMessage(['SendResponse - Read failed !!!!!', PAnsiChar(IntToStr(ss))], 0, 0, Cardinal(Self));
{$ENDIF}
        SSLDisconnect;
      end;
    end;
  end;
end;

procedure THttpSrvClientSSL.SSLDisconnect;
var
  dwType: DWORD;
  OutBufDesc: SecBufferDesc;
  LOutBuf: SecBuffer;
  SSPIFlags: DWORD;
  SSPIOutFlags: DWORD;
  Expiry: TimeStamp;
  SS: SECURITY_STATUS;
  WsaBuf: TWsaBuf;
begin
{$IFDEF LOGGING_ON}
  Server.LogMessage(['SSLDisconnect - TRY'], 0, 0, Cardinal(Self));
{$ENDIF}
  if 0 = InterlockedExchange(FSSLConnected, 0) then Exit;

  try
{$IFDEF LOGGING_ON}
    Server.LogMessage(['SSLDisconnect - IN'], 0, 0, Cardinal(Self));
{$ENDIF}
    dwType:= SCHANNEL_SHUTDOWN;

    LOutBuf.pvBuffer  := @dwType;
    LOutBuf.BufferType:= SECBUFFER_TOKEN;
    LOutBuf.cbBuffer  := sizeof(dwType);

    OutBufDesc.cBuffers := 1;
    OutBufDesc.pBuffers := @LOutBuf;
    OutBufDesc.ulVersion:= SECBUFFER_VERSION;

    SS:= SslSecurity.SSPI.ApplyControlToken(@FContext, @OutBufDesc);

    if FAILED(SS) then Exit;

    SSPIFlags:= ASC_REQ_SEQUENCE_DETECT or ASC_REQ_REPLAY_DETECT   or
                ASC_REQ_CONFIDENTIALITY or ASC_REQ_EXTENDED_ERROR  or
                ASC_REQ_ALLOCATE_MEMORY or ASC_REQ_STREAM;

    FOutBuffer.pvBuffer  := nil;
    FOutBuffer.BufferType:= SECBUFFER_TOKEN;
    FOutBuffer.cbBuffer  := 0;

    OutBufDesc.cBuffers := 1;
    OutBufDesc.pBuffers := @FOutBuffer;
    OutBufDesc.ulVersion:= SECBUFFER_VERSION;

    SS:= SslSecurity.SSPI.AcceptSecurityContext(@SslSecurity.FCredHandle,
                    @FContext, nil, SSPIFlags, SECURITY_NATIVE_DREP,
                    nil, @OutBufDesc, SSPIOutFlags, @Expiry);

    if FAILED(SS) then Exit;

    if (FOutBuffer.cbBuffer <> 0) and (FOutBuffer.pvBuffer <> nil) then
    begin
      WsaBuf.cLength:= FOutBuffer.cbBuffer;
      WsaBuf.pBuffer:= FOutBuffer.pvBuffer;
      Write(WsaBuf, 1, 0);
    end;

  finally
    FreeSecurityContext;
    FClientState:= SSL_NOTCONNECTED;
    Disconnect;
  end;
end;

end.
